# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

from pkg_resources import get_distribution, DistributionNotFound

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = "unknown"



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Raster layer resources
"""
from actinia_core.processing.actinia_processing.ephemeral.persistent_processing import (
    PersistentProcessing,
)
from actinia_core.core.request_parser import extract_glist_parameters
from actinia_core.core.common.exceptions import AsyncProcessError
from actinia_core.models.response_models import (
    StringListProcessingResultResponseModel,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class PersistentListLayers(PersistentProcessing):
    """List all map layers (raster, vector) of a specific mapset,
    dependent on the provided type.
    """

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)
        self.response_model_class = StringListProcessingResultResponseModel

    def _execute(self):
        self._setup()

        args, layer_type = self.data
        self.required_mapsets.append(self.target_mapset_name)

        options = extract_glist_parameters(args)

        pc = {"1": {"module": "g.list", "inputs": {}}}

        for key in options:
            pc["1"]["inputs"][key] = options[key]

        pc["1"]["inputs"]["mapset"] = self.target_mapset_name
        pc["1"]["inputs"]["type"] = layer_type

        self.skip_region_check = True
        process_list = self._validate_process_chain(
            skip_permission_check=True, process_chain=pc
        )
        self._create_temp_database(self.required_mapsets)
        self._create_grass_environment(
            grass_data_base=self.temp_grass_data_base,
            mapset_name=self.target_mapset_name,
        )

        self._execute_process_list(process_list)

        raster_layers_with_mapset = []
        raster_layers = self.module_output_log[0]["stdout"].split()

        for raster_layer in raster_layers:
            raster_layers_with_mapset.append(raster_layer.strip())

        self.module_results = raster_layers_with_mapset


class PersistentRemoveLayers(PersistentProcessing):
    """Remove layers in a mapset"""

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)

    def _execute(self):
        self._setup()

        args, layer_type = self.data
        self.required_mapsets.append(self.target_mapset_name)

        options = extract_glist_parameters(args)

        pc = {"1": {"module": "g.remove", "inputs": {}, "flags": "f"}}
        for key in options:
            pc["1"]["inputs"][key] = options[key]
        pc["1"]["inputs"]["type"] = layer_type

        self.skip_region_check = True
        process_list = self._validate_process_chain(
            skip_permission_check=True, process_chain=pc
        )
        self._create_temp_database(self.required_mapsets)
        self._check_lock_target_mapset()
        self._create_grass_environment(
            grass_data_base=self.temp_grass_data_base,
            mapset_name=self.target_mapset_name,
        )

        self._execute_process_list(process_list)

        if "WARNING: No data base element files found" in "\n".join(
            self.module_output_log[0]["stderr"]
        ):
            raise AsyncProcessError("<%s> layer not found" % layer_type)

        self.finish_message = "Successfully removed %s layers." % layer_type


class PersistentRenameLayers(PersistentProcessing):
    """Rename raster layers in a mapset"""

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)

    def _execute(self):
        self._setup()

        args, layer_type = self.data
        self.required_mapsets.append(self.target_mapset_name)

        # List format must be
        # [(a, a_new),(b, b_new),(c, c_new), ...]
        name_list = list()
        for old_name, new_name in self.request_data:
            name_list.append("%s,%s" % (old_name, new_name))
        name_string = ",".join(name_list)

        pc = {"1": {"module": "g.rename", "inputs": {layer_type: name_string}}}

        self.skip_region_check = True
        process_list = self._validate_process_chain(
            skip_permission_check=True, process_chain=pc
        )
        self._create_temp_database(self.required_mapsets)
        self._check_lock_target_mapset()
        self._create_grass_environment(
            grass_data_base=self.temp_grass_data_base,
            mapset_name=self.target_mapset_name,
        )

        self._execute_process_list(process_list)

        if "WARNING: " in "\n".join(self.module_output_log[0]["stderr"]):
            if "not found" in "\n".join(self.module_output_log[0]["stderr"]):
                raise AsyncProcessError("Error while renaming map layers")

        self.finish_message = "Successfully renamed %s layers." % layer_type



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
GRASS GIS environment initialization
"""

import os
import os.path
import subprocess
import tempfile
import shutil
import uuid
from .messages_logger import MessageLogger

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2023, Sören Gebbert, Anika Weinmann and mundialis GmbH & "
    "Co. KG"
)
__maintainer__ = "mundialis"


class GrassInitError(Exception):
    """Exception that is thrown in case of a an
    initialization error of the GRASS environment
    """

    def __init__(self, message):
        message = "%s:  %s" % (str(self.__class__.__name__), message)
        Exception.__init__(self, message)
        logger = MessageLogger()
        logger.error(message)


class ProcessLogging:
    """This class initiates the logging mechanism and is the base class for all
    other classes, which have information to be logged.
    """

    def __init__(self, config=None, user_id=None):
        self.logger = MessageLogger(config=config, user_id=user_id)

    def log_debug(self, message):
        """Write an info message into the logfile"""
        if message is not None and message != "":
            self.logger.debug(message)

    def log_info(self, message):
        """Write an info message into the logfile"""
        if message is not None and message != "":
            self.logger.info(message)

    def log_warning(self, message):
        """Write a warning message into the logfile"""
        if message is not None and message != "":
            self.logger.warning(message)

    def log_error(self, message):
        """Write an error message into the logfile"""
        if message is not None and message != "":
            self.logger.error(message)


class GrassEnvironment(ProcessLogging):
    """This class saves and sets grass environment variables"""

    def __init__(self):
        ProcessLogging.__init__(self)
        self.env = {
            "GISBASE": "",
            "GISRC": "",
            "LD_LIBRARY_PATH": "",
            "GRASS_ADDON_PATH": "",
            "GRASS_VERSION": "",
            "PYTHONPATH": "",
            "GRASS_MESSAGE_FORMAT": "plain",
            "GRASS_SKIP_MAPSET_OWNER_CHECK": "1",
            "GRASS_TGIS_RAISE_ON_ERROR": "1",
        }

    def set_grass_environment(
        self, gisrc_path, grass_gis_base, grass_addon_path
    ):
        """Set the grass environment variables

        Args:
            gisrc_path (str): Path to the GISRC file
            grass_gis_base (str): GRASS installation directory
            grass_addon_path (str): The path to GRASS addons

        """
        self.env["GIS_LOCK"] = str(os.getpid())
        self.env["GISBASE"] = grass_gis_base
        self.env["HOME"] = os.getenv("HOME", "/tmp/")
        self.env["GRASS_MESSAGE_FORMAT"] = "plain"
        self.env["GRASS_SKIP_MAPSET_OWNER_CHECK"] = "1"
        self.env["GRASS_TGIS_RAISE_ON_ERROR"] = "1"
        self.env["GISRC"] = os.path.join(gisrc_path, "gisrc")
        self.env["LD_LIBRARY_PATH"] = str(
            os.path.join(self.env["GISBASE"], "lib")
        )
        self.env["GRASS_VERSION"] = "8"
        self.env["GRASS_ADDON_PATH"] = grass_addon_path
        self.env["GRASS_ADDON_BASE"] = grass_addon_path
        if os.name != "posix":
            self.env["PATH"] = str(
                os.path.join(self.env["GISBASE"], "bin")
                + ";"
                + os.path.join(self.env["GISBASE"], "scripts")
                + ";"
                + os.path.join(self.env["GISBASE"], "lib")
                + ";"
                + os.path.join(self.env["GISBASE"], "extralib")
            )
            self.env["PYTHONPATH"] = str(
                self.env["PYTHONPATH"]
                + ";"
                + os.path.join(self.env["GISBASE"], "etc", "python")
            )
        else:
            self.env["PATH"] = str(
                os.path.join(self.env["GISBASE"], "bin")
                + ":"
                + os.path.join(self.env["GISBASE"], "scripts")
            )
            self.env["PYTHONPATH"] = str(
                self.env["PYTHONPATH"]
                + ":"
                + os.path.join(self.env["GISBASE"], "etc", "python")
            )

        self.set()
        self.get()

    def get(self):
        for key in self.env:
            try:
                self.env[key] = os.getenv(key, self.env[key])
            except Exception as e:
                raise GrassInitError(
                    "Error getting grass environmental variables. Exception: "
                    + str(e)
                )

    def set(self):
        for key in self.env:
            value = self.env[key]
            # use self.env and enviroment variable values
            if key in ["PATH", "PYTHONPATH"]:
                origValue = os.getenv(key, None)
                if origValue:
                    value += ":" + origValue
            try:
                os.putenv(key, value)
                os.environ[key] = value
                self.log_debug(key + "=" + value)
            except Exception as e:
                raise GrassInitError(
                    "Error setting grass environmental variables. Exception: "
                    + str(e)
                )


class GrassGisRC(ProcessLogging):
    """This class takes care of the correct creation of the gisrc file

    ATTENTION: Here the TGIS_DISABLE_MAPSET_CHECK variable is set True
    to allow map registering from none-current mapsets
    """

    def __init__(self, gisdbase, location_name, mapset):
        ProcessLogging.__init__(self)
        self.location_name = location_name
        self.mapset = mapset
        self.gisdbase = gisdbase
        self.__gisrc_ile = ""
        self.__tmpDir = ""

    def rewrite_file(self):
        if self.__gisrc_ile != "":
            self.__write()
        else:
            raise GrassInitError("Error re-writing the gisrc file")

    def write(self, gisrc_path):
        if os.path.isdir(gisrc_path):
            try:
                self.__gisrc_ile = os.path.join(gisrc_path, "gisrc")
                self.__write()
            except Exception:
                raise GrassInitError("Error writing the gisrc file")

    def print_gisrc(self):
        try:
            gisrc = open(self.__gisrc_ile, "r")
            self.log_debug(str(gisrc.read()))
            gisrc.close()
        except Exception:
            raise GrassInitError("Error printing the gisrc file")

    def __write(self):
        try:
            gisrc = open(self.__gisrc_ile, "w")
            gisrc.write("LOCATION_NAME: %s\n" % self.location_name)
            gisrc.write("MAPSET: %s\n" % self.mapset)
            gisrc.write("DIGITIZER: none\n")
            gisrc.write("GISDBASE: %s\n" % self.gisdbase)
            gisrc.write("OVERWRITE: 1\n")
            gisrc.write("DEBUG: 0\n")
            gisrc.write("GRASS_GUI: text")
            gisrc.write("TGIS_DISABLE_MAPSET_CHECK: 1")
            gisrc.close()
            gisrc = open(self.__gisrc_ile, "r")
            self.log_debug(gisrc.read())
            gisrc.close()
        except Exception:
            raise GrassInitError("Error writing the gisrc file")

    def get_filename(self):
        return self.__gisrc_ile


class GrassWindFile(ProcessLogging):
    """This class takes care of the correct creation of grass WIND and
    DEFAULT_WIND files using a dummy region"""

    def __init__(self, gisdbase, location, mapset):
        """

        Args:
            gisdbase (str): The GRASS database
            location (str): The location name
            mapset (str): The name of the mapset

        Raises:
            GrassInitError if unable to write wind file

        """
        ProcessLogging.__init__(self)
        """ Create the WIND and if needed the DEFAULT_WIND file """
        self.__windFile = ""
        self.__windname = "WIND"

        if mapset == "PERMANENT":
            # If PERMANENT is used as mapset, the DEFAULT_WIND file will be
            # created too
            self.__windFile = os.path.join(
                gisdbase, location, mapset, "DEFAULT_WIND"
            )
            self.__write()

        self.__windFile = os.path.join(gisdbase, location, mapset, "WIND")
        self.__write()

        try:
            wind = open(self.__windFile, "w")
            wind.write("""proj:       0\n""")
            wind.write("""zone:       0\n""")
            wind.write("""north:      100\n""")
            wind.write("""south:      0\n""")
            wind.write("""east:       100\n""")
            wind.write("""west:       0\n""")
            wind.write("""cols:       100\n""")
            wind.write("""rows:       100\n""")
            wind.write("""e-w resol:  1\n""")
            wind.write("""n-s resol:  1\n""")
            wind.close()
        except Exception:
            raise GrassInitError("Error writing the WIND file")

    def getFileName(self):
        return self.__windFile


class GrassModuleRunner(ProcessLogging):
    def __init__(self, grassbase, grass_addon_path):
        ProcessLogging.__init__(self)

        self.grassbase = grassbase
        self.grass_addon_path = grass_addon_path

    def _run_process(
        self,
        inputlist,
        raw=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
    ):
        """This function runs a process and logs its stdout and stderr output.
        It either returns the subprocess or its error id, stderr and stdout

        Args:
            inputlist (list): The input list for subprocess.Popen(args)
            raw (bool): If True return the subprocess, the caller has to take
                        care of it
            stdout (file): A file object that receives stdout, default
                           subprocess.PIPE
            stderr (file): A file object that receives stderr, default
                           subprocess.PIPE

        Returns:
            subprocess:
            The subprocess or a tuple of (errorid, stdout_buff, stderr_buff)

        """
        try:
            self.log_info("Run process: " + str(inputlist))
            proc = subprocess.Popen(
                args=inputlist, stdout=stdout, stderr=stderr, stdin=stdin
            )
            self.runPID = proc.pid
            self.log_debug("Process pid: " + str(self.runPID))

            if raw is True:
                return proc

            (stdout_buff, stderr_buff) = proc.communicate()
            stdout_buff = stdout_buff.decode()
            stderr_buff = stderr_buff.decode()

            self.log_debug("Return code: " + str(proc.returncode))
            self.log_debug(stderr_buff)
        except Exception:
            raise GrassInitError(
                "Unable to execute process: " + str(inputlist)
            )

        return proc.returncode, stdout_buff, stderr_buff

    def _create_grass_module_path(self, grass_module):
        """
        Create the parameter list and start the grass module. Search for grass
        modules in different grass specific directories

        Args:
            grass_module: The name of the module

        """

        if os.name != "posix":
            grass_module = grass_module + ".exe"

        pathList = []

        # Search the module in the bin directory
        grass_module_path = os.path.join(self.grassbase, "bin", grass_module)
        pathList.append(grass_module_path)
        self.log_debug("Looking for %s" % grass_module_path)

        if os.path.isfile(grass_module_path) is not True:
            grass_module_path = os.path.join(
                self.grassbase, "scripts", grass_module
            )
            pathList.append(grass_module_path)
            self.log_debug("Looking for %s" % grass_module_path)
            # if the module was not found in the bin dir, test the script
            # directory
            if os.path.isfile(grass_module_path) is not True:
                grass_module_path = os.path.join(
                    self.grass_addon_path, "bin", grass_module
                )
                pathList.append(grass_module_path)
                self.log_debug("Looking for %s" % grass_module_path)
                # if the module was not found in the script dir, test the addon
                # path
                if os.path.isfile(grass_module_path) is not True:
                    grass_module_path = os.path.join(
                        self.grass_addon_path, "scripts", grass_module
                    )
                    pathList.append(grass_module_path)
                    self.log_debug("Looking for %s" % grass_module_path)
                    if os.path.isfile(grass_module_path) is not True:
                        raise GrassInitError(
                            "GRASS module "
                            + grass_module
                            + " not found in "
                            + str(pathList)
                        )

        self.log_debug("GRASS module path is " + grass_module_path)

        return grass_module_path

    def run_module(
        self,
        grass_module,
        args,
        raw=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
    ):
        """Set all input and output options and start the module

        Raises:
            This method raises a GrassInitError Exception in case
            the GRASS module returns with an error. This happens only
            if raw=False.

        Args:
            grass_module (str): The module name, eg.: "g.region"
            args (list): The module arguments as list e.g.:
                         ["raster=slope", "-a"]
            raw (bool): If True the subprocess that run the module is returned,
                        otherwise a tuple of (errorid, stdout_buff,
                        stderr_buff) is returned
            stdout (file): A file object that receives stdout, default
                           subprocess.PIPE
            stderr (file): A file object that receives stderr, default
                           subprocess.PIPE
            stdin (file): A file object that provides stdin, default
                          subprocess.PIPE

        Returns:
            subprocess:
            Either the subprocess or a tuple of (errorid, stdout_buff,
            stderr_buff)

        """

        parameter = []
        grass_module_path = self._create_grass_module_path(grass_module)
        parameter.append(grass_module_path)
        parameter.extend(args)

        if raw is False:
            errorid, stdout_buff, stderr_buff = self._run_process(parameter)
        else:
            return self._run_process(
                parameter, raw=raw, stdout=stdout, stderr=stderr, stdin=stdin
            )

        if errorid != 0:
            log = (
                "Error while executing the grass module. "
                " \
            "
                "The following error message was logged:\n" + stderr_buff
            )
            raise GrassInitError(log)

        return errorid, stdout_buff, stderr_buff


class GrassInitializer(ProcessLogging):
    def __init__(
        self,
        grass_data_base,
        grass_base_dir,
        location_name,
        mapset_name,
        grass_addon_path="",
        config=None,
        user_id=None,
    ):
        """Initilialze the GRASS environment.

        Use an instance of this class for each independent process that should
        run GRASS modules or script to avoid conflicts. This class will create
        an environment to run GRASS GIS modules and Python scripts.

        Args:
            grass_data_base (str): GRASS GIS database root directory that
                                   contains locations
            grass_base_dir (str): The installation directory of GRASS GIS
            location_name (str): The name of the location to work in
            mapset_name (str): The name of the mapset to work in
            grass_addon_path (str): The path to GRASS GIS addons

        """

        ProcessLogging.__init__(self, config=config, user_id=user_id)

        self.gisrc_path = None
        self.grass_data_base = grass_data_base
        self.grass_base_dir = grass_base_dir
        self.location_name = location_name
        self.mapset_name = mapset_name
        self.grass_addon_path = grass_addon_path
        self.has_temp_region = False

    def initialize(self):
        """
        Initialize the GRASS environment to run GRASS commands
        """

        self.gisrc_path = tempfile.mkdtemp()

        self.mapset_path = os.path.join(
            self.grass_data_base, self.location_name, self.mapset_name
        )

        # Generate a temporary region name
        self.tmp_region_name = "tmp." + str(uuid.uuid1())

        self.genv = GrassEnvironment()
        self.genv.set_grass_environment(
            self.gisrc_path, self.grass_base_dir, self.grass_addon_path
        )

        self.gisrc = GrassGisRC(
            self.grass_data_base, self.location_name, self.mapset_name
        )
        self.gisrc.write(self.gisrc_path)

        self.runner = GrassModuleRunner(
            self.grass_base_dir, self.grass_addon_path
        )

    def run_module(
        self,
        module_name,
        parameter_list,
        raw=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
    ):
        """Run a grass module

        Args:
            module_name (str): The name of the module to run
            parameter_list (list): A list of parameter for the module
            raw (bool): If True the subprocess that run the module is returned,
                        otherwise a tuple of (errorid, stdout_buff,
                        stderr_buff) is returned
            stdout (file): A file object that receives stdout, default
                           subprocess.PIPE
            stderr (file): A file object that receives stderr, default
                           subprocess.PIPE

        Raises:
            This method raises a GrassInitError Exception in case
            the GRASS module returns with an error. This happens only
            if raw=False.

        Returns:
            subprocess:
            Either the subprocess or a tuple of (errorid, stdout_buff,
            stderr_buff)

        """
        return self.runner.run_module(
            module_name,
            parameter_list,
            raw,
            stdout=stdout,
            stderr=stderr,
            stdin=stdin,
        )

    def clean_up(self):
        """Try to remove the temporary gisrc file and the mapset lock"""
        self.delete_tmp_region()
        if self.gisrc_path is not None and os.path.isdir(self.gisrc_path):
            shutil.rmtree(self.gisrc_path)
        else:
            logger = MessageLogger()
            logger.error(
                "Unable to delete temporary grass database <%s>"
                % self.gisrc_path
            )

    def setup_tmp_region(self):
        """Setup a temporary region, so that g.region calls can be performed
        without altering the mapsets current region.

        Make sure to call delete_tmp_region() in the same process after
        processing finished, to make sure that the temporary region was removed
        and the environmental variable was unset.

        Raises:
            This method raises a GrassInitError Exception

        """
        # Safe the current region in a temporary region that can be overwritten
        errorid, stdout_buff, stderr_buff = self.run_module(
            "g.region", ["save=%s" % self.tmp_region_name, "--o"]
        )

        if errorid != 0:
            raise GrassInitError("Unable to create a temporary region")

        # Put the region in the process environment
        os.environ["WIND_OVERRIDE"] = self.tmp_region_name
        self.has_temp_region = True

    def delete_tmp_region(self):
        """Remove the temporary region is exists

        This method will only try to remove a temporary
        region that were created with this object

        """
        if self.has_temp_region is True:
            try:
                if "WIND_OVERRIDE" in os.environ:
                    os.environ.pop("WIND_OVERRIDE")
                    self.run_module(
                        "g.remove",
                        [
                            "name=%s" % self.tmp_region_name,
                            "type=region",
                            "-f",
                        ],
                    )
            except Exception:
                logger = MessageLogger()
                logger.error(
                    "Unable to delete temporary region <%s>"
                    % self.tmp_region_name
                )

            self.has_temp_region = False



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Raster map renderer
"""
from tempfile import NamedTemporaryFile
from actinia_core.processing.actinia_processing.ephemeral_processing import (
    EphemeralProcessing,
)
from actinia_core.processing.actinia_processing.ephemeral.base.renderer_base import (
    EphemeralRendererBase,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class EphemeralSTRDSRenderer(EphemeralRendererBase):
    def __init__(self, *args):
        EphemeralProcessing.__init__(self, *args)

    def _execute(self, skip_permission_check=True):
        """Render the raster map layers of a STRDS

        Workflow:

            1. A list of raster map layers is generated from a t.rast.list call
               that can be constrained with time and region settings
            1. The default region is set to the the cumulative raster layer
               regions
            2. User specific region settings are applied
            3. d.rast.multi is invoked to create the PNG file

        """

        self._setup()

        strds_name = self.map_name
        options = self.rdc.user_data
        self.required_mapsets.append(self.mapset_name)

        with NamedTemporaryFile(suffix=".png") as file:
            result_file = file.name

        g_region_query = self._setup_render_environment_and_region(
            options=options, result_file=result_file, legacy=False
        )
        where_list = []

        if "start_time" in options:
            where_list.append("start_time >= '%s'" % options["start_time"])
        if "end_time" in options:
            where_list.append("end_time  <= '%s'" % options["end_time"])
        if "n" in options:
            where_list.append("south <= %f" % options["n"])
        if "s" in options:
            where_list.append("north >= %f" % options["s"])
        if "e" in options:
            where_list.append("west <= %f" % options["e"])
        if "w" in options:
            where_list.append("east >= %f" % options["w"])

        where = " AND ".join(where_list)

        t_rast_list = {
            "id": "1",
            "module": "t.rast.list",
            "inputs": [
                {
                    "param": "input",
                    "value": "%s@%s" % (strds_name, self.mapset_name),
                },
                {"param": "method", "value": "comma"},
                {"param": "separator", "value": "comma"},
                {"param": "columns", "value": "id"},
                {"param": "format", "value": "line"},
                {"param": "where", "value": where},
            ],
            "flags": "u",
        }

        pc = {"version": 1, "list": []}
        pc["list"].append(t_rast_list)
        # Run the selected modules
        self.skip_region_check = True
        process_chain = (
            self._create_temporary_grass_environment_and_process_list(
                process_chain=pc, skip_permission_check=True
            )
        )
        self._execute_process_list(process_chain)

        map_list = self.module_output_log[0]["stdout"].strip()

        g_region = {
            "id": "2",
            "module": "g.region",
            "inputs": [{"param": "raster", "value": map_list}],
            "flags": "p",
        }

        g_region_query["id"] = "3"
        g_region_query["flags"] = "g"

        pc = {"version": 1, "list": []}
        pc["list"].append(g_region)
        pc["list"].append(g_region_query)

        process_list = self._validate_process_chain(
            process_chain=pc, skip_permission_check=True
        )
        self._execute_process_list(process_list)

        # Compute the cell size for visualization
        region_settings = self.module_output_log[2]["stdout"].strip()
        param_list = region_settings.split()
        n = 0
        s = 0
        e = 0
        w = 0
        for param in param_list:
            if "n=" == param[:2]:
                n = float(param.split("=")[1])
            if "s=" == param[:2]:
                s = float(param.split("=")[1])
            if "e=" == param[:2]:
                e = float(param.split("=")[1])
            if "w=" == param[:2]:
                w = float(param.split("=")[1])

        width = float(options["width"])
        height = float(options["height"])

        ewres = abs(e - w) / width
        nsres = abs(n - s) / height

        g_region_adjust = {
            "id": "4",
            "module": "g.region",
            "inputs": [
                {"param": "ewres", "value": "%f" % ewres},
                {"param": "nsres", "value": "%f" % nsres},
            ],
            "flags": "g",
        }

        d_rast = {
            "id": "6",
            "module": "d.rast.multi",
            "inputs": [{"param": "map", "value": map_list}],
        }

        pc = {"version": 1, "list": []}
        pc["list"].append(g_region_adjust)
        pc["list"].append(d_rast)

        process_list = self._validate_process_chain(
            process_chain=pc, skip_permission_check=True
        )
        self._execute_process_list(process_list)

        self.module_results = result_file



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Tests: Raster async export test case
"""
import unittest
import time

try:
    from .test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX
except ModuleNotFoundError:
    from test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class RasterAsyncExport(ActiniaResourceTestCaseBase):
    def test_export(self):
        rv = self.server.post(
            f"{URL_PREFIX}/locations/nc_spm_08/mapsets/PERMANENT/raster_layers"
            "/elevation/geotiff_async",
            headers=self.user_auth_header,
        )
        resp = self.waitAsyncStatusAssertHTTP(
            rv, headers=self.user_auth_header
        )

        # Get the exported results
        urls = resp["urls"]["resources"]

        for url in urls:
            print(url)
            rv = self.server.get(url, headers=self.user_auth_header)
            self.assertEqual(
                rv.status_code,
                200,
                "HTML status code is wrong %i" % rv.status_code,
            )
            self.assertEqual(
                rv.mimetype, "image/tiff", "Wrong mimetype %s" % rv.mimetype
            )
            print(rv.headers)

        time.sleep(1)

    def test_export_region(self):
        rv = self.server.post(
            f"{URL_PREFIX}/locations/nc_spm_08/mapsets/PERMANENT/raster_layers"
            "/elevation/geotiff_async_orig",
            headers=self.user_auth_header,
        )
        resp = self.waitAsyncStatusAssertHTTP(
            rv, headers=self.user_auth_header
        )

        # Get the exported results
        urls = resp["urls"]["resources"]

        for url in urls:
            print(url)
            rv = self.server.get(url, headers=self.user_auth_header)
            self.assertEqual(
                rv.status_code,
                200,
                "HTML status code is wrong %i" % rv.status_code,
            )
            self.assertEqual(
                rv.mimetype, "image/tiff", "Wrong mimetype %s" % rv.mimetype
            )
            print(rv.headers)

        time.sleep(1)

    def test_export_error(self):
        rv = self.server.post(
            f"{URL_PREFIX}/locations/nc_spm_08/mapsets/PERMANENT/"
            "raster_layers/elevationion/geotiff_async",
            headers=self.user_auth_header,
        )
        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.user_auth_header,
            http_status=400,
            status="error",
            message_check="AsyncProcessError:",
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Storage base class
"""
from abc import ABCMeta, abstractmethod

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class ResourceStorageBase(object):
    """Abstract storage base class of generated resources"""

    __metaclass__ = ABCMeta

    def __init__(self, user_id, resource_id, config):
        """Abstract storage base class of generated resources

        Args:
            user_id: The user id
            resource_id: The resource id
            config: The configuration of actinia core
        """

        self.user_id = user_id
        self.resource_id = resource_id
        self.config = config
        self.resource_url_list = []
        self.resource_file_list = []

    @abstractmethod
    def setup(self):
        """Setup the resource storage"""
        pass

    @abstractmethod
    def get_resource_urls(self):
        """Return all resource urls that were generated when storing a resource
        on disk

        Returns:
            (list): A list of urls

        """
        return self.resource_url_list

    @abstractmethod
    def store_resource(self, file_path):
        """Store a resource (file) at the user resource storage and return an
        URL to the resource accessible via HTTP

        Args:
            file_path:

        Returns:
            (str): the resource url that points to the stored resource

        """
        pass

    @abstractmethod
    def remove_resources(self):
        """Remove all stored resources"""
        pass



#!flask/bin/python
# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Actinia Core
"""

import os
from .endpoints import create_endpoints
from .health_check import health_check
from .version import version, init_versions
from actinia_core.core.common.app import flask_app
from actinia_core.core.common.config import global_config, DEFAULT_CONFIG_PATH
from actinia_core.core.common.redis_interface import connect
from actinia_core.core.common.process_queue import create_process_queue

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"

# if os.environ.get('DEFAULT_CONFIG_PATH'):
#     DEFAULT_CONFIG_PATH = os.environ['DEFAULT_CONFIG_PATH']
if os.path.exists(DEFAULT_CONFIG_PATH) is True and os.path.isfile(
    DEFAULT_CONFIG_PATH
):
    global_config.read(DEFAULT_CONFIG_PATH)

# Create the endpoints based on the global config
create_endpoints()
init_versions()

# TODO: Implement a better error handler
# @flask_app.errorhandler(InvalidUsage)
# def handle_invalid_usage(error):
#    response = error.to_json()
#    response.status_code = error.status_code
#    return response

# Connect the redis interfaces
redis_args = (global_config.REDIS_SERVER_URL, global_config.REDIS_SERVER_PORT)
if global_config.REDIS_SERVER_PW and global_config.REDIS_SERVER_PW is not None:
    redis_args = (*redis_args, global_config.REDIS_SERVER_PW)

connect(*redis_args)
del redis_args

# Create the process queue
create_process_queue(global_config)

###############################################################################
if __name__ == "__main__":
    # Connect to the database
    flask_app.run(host="0.0.0.0", port=8080, debug=True)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Tests: STRDS render test case
"""
import unittest
from flask.json import dumps as json_dumps

try:
    from .test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX
except ModuleNotFoundError:
    from test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX


__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Soeren Gebbert"
__email__ = "soerengebbert@googlemail.com"

location = "nc_spm_08"
strds_mapset = "modis_lst"
strds_url = URL_PREFIX + "/locations/%(location)s/mapsets/%(mapset)s/strds" % {
    "location": location,
    "mapset": strds_mapset,
}
strds_data = "LST_Day_monthly"


class STRDSRenderTestCase(ActiniaResourceTestCaseBase):
    def create_raster_layer(
        self, location_name, mapset_name, raster_name, val
    ):
        # Remove potentially existing raster layer
        rv = self.server.delete(
            URL_PREFIX
            + "/locations/%s/mapsets/%s/raster_layers/%s"
            % (location_name, mapset_name, raster_name),
            headers=self.admin_auth_header,
        )
        # print(rv.data)
        # Create
        postbody = {
            "list": [
                {
                    "id": "set_region",
                    "module": "g.region",
                    "inputs": [
                        {"param": "n", "value": "228500"},
                        {"param": "s", "value": "215000"},
                        {"param": "e", "value": "645000"},
                        {"param": "w", "value": "630000"},
                        {"param": "ewres", "value": "50"},
                        {"param": "nsres", "value": "50"},
                    ],
                },
                {
                    "id": "create_raster",
                    "module": "r.mapcalc",
                    "inputs": [
                        {
                            "param": "expression",
                            "value": "%s = %d" % (raster_name, val),
                        }
                    ],
                },
            ],
            "version": "1",
        }
        rv = self.server.post(
            URL_PREFIX
            + "/locations/%s/mapsets/%s/processing_async"
            % (location_name, mapset_name),
            headers=self.admin_auth_header,
            data=json_dumps(postbody),
            content_type="application/json",
        )
        self.waitAsyncStatusAssertHTTP(
            rv,
            headers=self.admin_auth_header,
            http_status=200,
            status="finished",
        )
        # print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

    def test_strds_render_1(self):
        new_mapset = "strds_render_test"
        self.create_new_mapset(new_mapset, location)

        # Create success
        rv = self.server.post(
            f"{URL_PREFIX}/locations/{location}/mapsets/{new_mapset}/strds/"
            "test_strds_register",
            headers=self.admin_auth_header,
            data=json_dumps(
                {
                    "temporaltype": "absolute",
                    "title": "A nice title",
                    "description": "A nice description",
                }
            ),
            content_type="application/json",
        )

        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # Create the raster layer
        self.create_raster_layer(location, new_mapset, "test_layer_1", 1)
        self.create_raster_layer(location, new_mapset, "test_layer_2", 2)
        self.create_raster_layer(location, new_mapset, "test_layer_3", 3)

        raster_layers = [
            {
                "name": "test_layer_1",
                "start_time": "2000-01-01",
                "end_time": "2000-01-02",
            },
            {
                "name": "test_layer_2",
                "start_time": "2000-01-02",
                "end_time": "2000-01-03",
            },
            {
                "name": "test_layer_3",
                "start_time": "2000-01-03",
                "end_time": "2000-01-04",
            },
        ]

        rv = self.server.put(
            f"{URL_PREFIX}/locations/{location}/mapsets/{new_mapset}/strds/"
            "test_strds_register/raster_layers",
            data=json_dumps(raster_layers),
            content_type="application/json",
            headers=self.admin_auth_header,
        )

        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "application/json", "Wrong mimetype %s" % rv.mimetype
        )

        # Check strds
        rv = self.server.get(
            f"{URL_PREFIX}/locations/{location}/mapsets/{new_mapset}/strds/"
            "test_strds_register/render?width=100&height=100",
            headers=self.admin_auth_header,
        )

        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "image/png", "Wrong mimetype %s" % rv.mimetype
        )

        # Check strds
        rv = self.server.get(
            f"{URL_PREFIX}/locations/{location}/mapsets/{new_mapset}/strds/"
            "test_strds_register/render?width=100&height=100&"
            "start_time=2000-01-01 00:00:00&end_time=2000-01-02 00:00:00",
            headers=self.admin_auth_header,
        )
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )
        self.assertEqual(
            rv.mimetype, "image/png", "Wrong mimetype %s" % rv.mimetype
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Tests: version health
"""
import unittest

try:
    from .test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX
except ModuleNotFoundError:
    from test_resource_base import ActiniaResourceTestCaseBase, URL_PREFIX


__license__ = "GPLv3"
__author__ = "Sören Gebbert, Julia Haas"
__copyright__ = (
    "Copyright 2016-2021, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class VersionHealthTestCase(ActiniaResourceTestCaseBase):
    def test_version(self):
        """test version endpoint"""
        rv = self.server.get(URL_PREFIX + "/version")
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )

    def test_version_slash(self):
        """test version endpoint with trailing slash"""
        rv = self.server.get(URL_PREFIX + "/version/")
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )

    def test_health_check(self):
        """test health_check endpoint"""
        rv = self.server.get(URL_PREFIX + "/health_check")
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )

    def test_health_check_slash(self):
        """test health_check endpoint with trailing slash"""
        rv = self.server.get(URL_PREFIX + "/health_check/")
        print(rv.data)
        self.assertEqual(
            rv.status_code,
            200,
            "HTML status code is wrong %i" % rv.status_code,
        )


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Sentinel-2A processing commands
"""
import os
import requests
import dateutil.parser as dtparser
from .google_satellite_bigquery_interface import (
    GoogleSatelliteBigQueryInterface,
)
from .aws_sentinel_interface import AWSSentinel2AInterface
from .exceptions import AsyncProcessError
from .process_object import Process


__license__ = "GPLv3"
__author__ = "Sören Gebbert, Guido Riembauer"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis GmbH & Co. KG"


def datetime_to_grass_datetime_string(dt):
    """Convert a python datetime object into a GRASS datetime string"""
    # GRASS datetime month names
    month_names = [
        "",
        "jan",
        "feb",
        "mar",
        "apr",
        "may",
        "jun",
        "jul",
        "aug",
        "sep",
        "oct",
        "nov",
        "dec",
    ]

    # Check for time zone info in the datetime object
    if dt.tzinfo is not None:
        tz = dt.tzinfo.utcoffset(0)
        if tz.seconds > 86400 / 2:
            tz = (tz.seconds - 86400) / 60
        else:
            tz = tz.seconds / 60

        string = "%.2i %s %.2i %.2i:%.2i:%.2i %+.4i" % (
            dt.day,
            month_names[dt.month],
            dt.year,
            dt.hour,
            dt.minute,
            dt.second,
            tz,
        )
    else:
        string = "%.2i %s %.4i %.2i:%.2i:%.2i" % (
            dt.day,
            month_names[dt.month],
            dt.year,
            dt.hour,
            dt.minute,
            dt.second,
        )

    return string


class Sentinel2Processing(object):
    """"""

    def __init__(
        self,
        product_id,
        bands,
        download_cache,
        send_resource_update,
        message_logger,
        use_google=True,
        temp_file_path=None,
        config=None,
        query_result=None,
    ):
        """A collection of functions to generate Sentinel2 related import and
        processing commands. Each function returns a process chain that can be
        executed by the async processing classes.

        Args:
            config: The Actinia Core configuration object
            product_id (str): The scene id for which all bands should be
                              downloaded
            query_result (dict): The result of the BigQuery request
            bands (list): A list of band names
            temp_file_path: The path to the temporary directory to store
                            temporary files. It is assumed that this path is
                            available when the generated commands are executed.
            download_cache (str): The path to the download cache
            send_resource_update: The function to call for resource updates
            message_logger: The message logger to be used

        """
        self.config = config
        self.product_id = product_id
        self.query_result = query_result
        self.bands = bands
        self.temp_file_path = temp_file_path
        self.user_download_cache_path = download_cache
        self.send_resource_update = send_resource_update
        self.message_logger = message_logger
        self.gml_cache_file_name = None
        self.import_file_info = {}
        self.gml_temp_file_name = None
        self.timestamp = None
        self.bbox = None
        self.band_pattern = None
        self.use_google = use_google

    def _setup_download_import_google_without_query(self):
        """Setup the map name info for later renaming and the band pattern for
        importing.
        """
        level = self.product_id.split("_")[1]
        tile_block = self.product_id.split("_")[-2]
        date_block = self.product_id.split("_")[2]
        band_pattern = "("
        # Create file names, urls and check the download cache
        for band in self.bands:
            # kept here only for consistency:
            file_path = None
            # build the map name, here the highest possible resolution for each
            # band is assumed
            if band in ["B02", "B03", "B04", "B08"]:
                res = "10"
            elif band in ["B05", "B06", "B07", "B8A", "B11", "B12"]:
                res = "20"
            elif band in ["B1", "B9", "B10"]:
                res = "60"
            else:
                raise AsyncProcessError(
                    f"Band {band} is unknown. Please "
                    "provide band as 'BXY', e.g. "
                    "'B02', 'B8A', or 'B12'"
                )
            if level == "MSIL1C":
                band_pattern += f"{band}|"
                map_name = f"{tile_block}_{date_block}_{band}"
            elif level == "MSIL2A":
                band_pattern += f"{band}_{res}m|"
                map_name = f"{tile_block}_{date_block}_{band}_{res}m"
            self.import_file_info[band] = (file_path, map_name)
        band_pattern = band_pattern[:-1] + ")"
        self.band_pattern = band_pattern

    def _setup_download_import_google(self):
        """Setup the download, import and preprocessing of a
        sentinel2 scene from the Google Cloud Storage.

        1. Query Google BogQuery to gather the scene information
        2. Create URL list of files that must be downloaded and copied to the
           download cache
        3. Create the GML file that represents the footprint of the scene
        4. Check if the requested files exist

        Returns:
            (url_list, copy_file_list)
        """

        if self.query_result is None:
            self.query_interface = GoogleSatelliteBigQueryInterface(
                self.config
            )
            try:
                self.query_result = self.query_interface.get_sentinel_urls(
                    [
                        self.product_id,
                    ],
                    self.bands,
                )
            except Exception as e:
                raise AsyncProcessError(
                    "Error in querying Sentinel-2 product <%s> "
                    "in Google BigQuery Sentinel-2 database. "
                    "Error: %s" % (self.product_id, str(e))
                )

            if not self.query_result:
                raise AsyncProcessError(
                    "Unable to find Sentinel-2 product <%s> "
                    "in Google BigQuery Sentinel-2 database" % self.product_id
                )

        if self.product_id not in self.query_result:
            raise AsyncProcessError(
                "Unable to find Sentinel-2 product <%s> "
                "in Google BigQuery Sentinel-2 database" % self.product_id
            )

        # Switch into the tempfile directory
        os.chdir(self.temp_file_path)

        url_list = []
        copy_file_list = []

        gml_footprint = self.query_result[self.product_id]["gml_footprint"]
        self.timestamp = self.query_result[self.product_id]["timestamp"]
        self.bbox = self.query_result[self.product_id]["bbox"]

        # Write the gml footprint into a temporary file
        # The file name is the product id that will also be copied into the
        # download cache
        self.gml_cache_file_name = os.path.join(
            self.user_download_cache_path, self.product_id + ".gml"
        )

        self.import_file_info["footprint"] = (
            self.gml_cache_file_name,
            self.product_id,
        )

        if os.path.exists(self.gml_cache_file_name) is False:
            gml_temp_file_name = os.path.join(
                self.temp_file_path, self.product_id + ".gml"
            )
            gml_file = open(gml_temp_file_name, "w")
            gml_file.write(gml_footprint)
            gml_file.close()
            copy_file_list.append(
                (gml_temp_file_name, self.gml_cache_file_name)
            )
            self.message_logger.info(
                (gml_temp_file_name, self.gml_cache_file_name)
            )

        # Create file names, urls and check the download cache
        for band in self.bands:
            file_name = self.query_result[self.product_id][band]["file"]
            tile_name = self.query_result[self.product_id][band]["tile"]
            public_url = self.query_result[self.product_id][band]["public_url"]

            # This is the file path in the download cache directory
            file_path = os.path.join(self.user_download_cache_path, file_name)

            self.import_file_info[band] = (file_path, file_name)

            # This is the download path
            temp_file = os.path.join(self.temp_file_path, tile_name)

            # Check if the file already exists in the download cache
            # if not add the file to the download list and copy list
            if os.path.exists(file_path) is False:
                self.message_logger.info(public_url)
                url_list.append(public_url)
                copy_file_list.append((temp_file, file_path))

        # Check the urls for access. If all files are already in the download
        # cache, then nothing needs to be downloaded and checked.
        for url in url_list:
            # Send a resource update
            self.send_resource_update(
                message="Checking access to URL: %s" % url
            )

            # Check if thr URL exists by investigating the HTTP header
            resp = requests.head(url)
            self.message_logger.info(
                "%i %s %s" % (resp.status_code, resp.text, resp.headers)
            )

            if resp.status_code != 200:
                raise AsyncProcessError(
                    "Scene <%s> is not available. "
                    "The URL <%s> can not be accessed."
                    % (self.product_id, url)
                )

        return url_list, copy_file_list

    def _setup_download_import_aws(self):
        """Setup the download, import and preprocessing of a
        sentinel2 scene from the AWS Storage.

        Returns:
            (url_list, copy_file_list)

        TODO:  The implementation is still in progress, thr AWS download does
               not word yet.
        """

        if self.query_result is None:
            self.query_interface = AWSSentinel2AInterface(self.config)
            try:
                self.query_result = self.query_interface.get_sentinel_urls(
                    [
                        self.product_id,
                    ],
                    self.bands,
                )
            except Exception as e:
                raise AsyncProcessError(
                    "Error in querying Sentinel-2 product <%s> "
                    "in AWS Sentinel-2 database. "
                    "Error: %s" % (self.product_id, str(e))
                )

            if not self.query_result:
                raise AsyncProcessError(
                    "Unable to find Sentinel-2 product <%s> "
                    "in AWS Sentinel-2 database" % self.product_id
                )

        if self.product_id not in self.query_result:
            raise AsyncProcessError(
                "Unable to find Sentinel-2 product <%s> "
                "in AWS Sentinel-2 database" % self.product_id
            )

        # Switch into the tempfile directory
        os.chdir(self.temp_file_path)

        url_list = []
        copy_file_list = []

        gml_footprint = self.query_result[self.product_id]["gml_footprint"]
        self.timestamp = self.query_result[self.product_id]["timestamp"]
        self.bbox = self.query_result[self.product_id]["bbox"]

        # Write the gml footprint into a temporary file
        # The file name is the product id that will also be copied into the
        # download cache
        self.gml_cache_file_name = os.path.join(
            self.user_download_cache_path, self.product_id + ".gml"
        )

        self.import_file_info["footprint"] = (
            self.gml_cache_file_name,
            self.product_id,
        )

        if os.path.exists(self.gml_cache_file_name) is False:
            gml_temp_file_name = os.path.join(
                self.temp_file_path, self.product_id + ".gml"
            )
            gml_file = open(gml_temp_file_name, "w")
            gml_file.write(gml_footprint)
            gml_file.close()
            copy_file_list.append(
                (gml_temp_file_name, self.gml_cache_file_name)
            )
            self.message_logger.info(
                (gml_temp_file_name, self.gml_cache_file_name)
            )

        # Create file names, urls and check the download cache
        for band in self.bands:
            file_name = self.query_result[self.product_id][band]["file"]
            tile_name = self.query_result[self.product_id][band]["tile"]
            public_url = self.query_result[self.product_id][band]["public_url"]

            # This is the file path in the download cache directory
            file_path = os.path.join(self.user_download_cache_path, file_name)

            self.import_file_info[band] = (file_path, file_name)

            # This is the download path
            temp_file = os.path.join(self.temp_file_path, tile_name)

            # Check if the file already exists in the download cache
            # if not add the file to the download list and copy list
            if os.path.exists(file_path) is False:
                self.message_logger.info(public_url)
                url_list.append(public_url)
                copy_file_list.append((temp_file, file_path))

        # Check the urls for access. If all files are already in the download
        # cache, then nothing needs to be downloaded and checked.
        for url in url_list:
            # Send a resource update
            self.send_resource_update(
                message="Checking access to URL: %s" % url
            )

            # Check if thr URL exists by investigating the HTTP header
            resp = requests.head(url)
            self.message_logger.info(
                "%i %s %s" % (resp.status_code, resp.text, resp.headers)
            )

            if resp.status_code != 200:
                raise AsyncProcessError(
                    "Scene <%s> is not available. "
                    "The URL <%s> can not be accessed."
                    % (self.product_id, url)
                )

        return url_list, copy_file_list

    def _setup_download_import(self):
        if self.use_google:
            return self._setup_download_import_google()
        else:
            return self._setup_download_import_aws()

    def get_sentinel2_download_process_list_without_query(self):
        """Create the process list to download sentinel2 scenes
        from the Google Cloud Storage.
        """
        download_commands = []

        p = Process(
            exec_type="grass",
            executable="i.sentinel.download",
            executable_params=[
                "datasource=GCS",
                f"query=identifier={self.product_id}",
                f"output={self.user_download_cache_path}",
            ],
            id=f"i_sentinel_download_{self.product_id}",
            skip_permission_check=True,
        )
        download_commands.append(p)

        # The renaming step needs to know the map name(s). They can be put
        # together from the filename
        self._setup_download_import_google_without_query()

        return download_commands, self.import_file_info

    def get_sentinel2_import_process_list_without_query(self):
        """Generate Sentinel2A import process list

        1. Import required bands using i.sentinel.import
        """
        import_commands = []

        p = Process(
            exec_type="grass",
            executable="i.sentinel.import",
            executable_params=[
                f"input={self.user_download_cache_path}",
                f"pattern={self.band_pattern}",
                "-r",
            ],
            id=f"i_sentinel_import_{self.product_id}",
            skip_permission_check=True,
        )
        import_commands.append(p)
        return import_commands

    def get_sentinel2_download_process_list(self):
        """Create the process list to download, import and preprocess
        sentinel2 scene from the Google Cloud Storage.

        The downloaded files will be stored in a temporary directory. After the
        download of all files completes, the downloaded files will be moved to
        the download cache. This avoids broken files in case a download was
        interrupted or stopped by termination.

        This method creates wget calls to gather the sentinel2 scenes from the
        Google Cloud Storage sentinel2 archive using public https address.

        Returns:
            (import_commands, import_file_info)

            Example:

                import_file_info = {"footprint": (gml_file_name, product_id),
                               "B01":(file_path, file_name),
                               "B02":(file_path, file_name)}
        """
        url_list, copy_file_list = self._setup_download_import()

        download_commands = []

        # Create the download commands and update process chain
        for url in url_list:
            wget = "/usr/bin/wget"
            wget_params = list()
            wget_params.append("-t5")
            wget_params.append("-c")
            wget_params.append("-q")
            wget_params.append(url)

            p = Process(
                exec_type="exec",
                executable=wget,
                executable_params=wget_params,
                id=f"wget_{url.split('/')[-1]}",
                skip_permission_check=True,
            )

            download_commands.append(p)

        # Create the commands to move the downloaded files to the download
        # cache
        for source, dest in copy_file_list:
            if source != dest:
                copy = "/bin/mv"
                copy_params = list()
                copy_params.append(source)
                copy_params.append(dest)

                p = Process(
                    exec_type="exec",
                    executable=copy,
                    executable_params=copy_params,
                    id=f"mv_{os.path.basename(dest)}",
                    skip_permission_check=True,
                )
                download_commands.append(p)

        return download_commands, self.import_file_info

    def get_sentinel2_import_process_list(self):
        """Generate Senteinel2A import and preprocessing process list

        1. Import footprint with v.import
        2. Set the timestamp of the footprint with v.timestamp

        For each band:

            0. Use gdaltrans to select the footprint bbox
               that should be imported from the raster layer
            1. Import band with r.import
            2. Use g.region to set the region to footprint
            3. Create a mask with r.mask
            4. Compute the cropped version of the band with r.mapcalc
            5. Set the timestamp with r.timestamp
            6. Remove uncropped version with g.remove
            7. Remove the mask with r.mask

        Returns:
            list[Process]:
            The list of import commands

        """

        import_commands = []

        p = Process(
            exec_type="grass",
            executable="v.import",
            executable_params=[
                "input=%s" % self.gml_cache_file_name,
                "output=%s" % self.product_id,
                "--q",
            ],
            id=f"v_import_{self.product_id}",
            skip_permission_check=True,
        )
        import_commands.append(p)

        dt = dtparser.parse(self.timestamp.split(".")[0])
        timestamp = datetime_to_grass_datetime_string(dt)

        # Attach a the time stamp
        p = Process(
            exec_type="grass",
            executable="v.timestamp",
            executable_params=[
                "map=%s" % self.product_id,
                "date=%s" % timestamp,
            ],
            id=f"v_timestamp_{self.product_id}",
            skip_permission_check=True,
        )
        import_commands.append(p)

        # Import and update
        for key in self.import_file_info:
            if key == "footprint":
                continue
            input_file, map_name = self.import_file_info[key]
            temp_map_name = map_name + "_uncropped"
            cropped_input_file = input_file + ".vrt"

            # Create a boundingbox around the footprint to avoid
            # the projection of the scene with unused values
            gdal_translate = "/usr/bin/gdal_translate"
            gdal_translate_params = list()
            # -projwin ulx uly lrx lry
            gdal_translate_params.append("-projwin")
            gdal_translate_params.append("%f" % self.bbox[0])
            gdal_translate_params.append("%f" % self.bbox[1])
            gdal_translate_params.append("%f" % self.bbox[2])
            gdal_translate_params.append("%f" % self.bbox[3])
            gdal_translate_params.append("-of")
            gdal_translate_params.append("vrt")
            gdal_translate_params.append("-projwin_srs")
            gdal_translate_params.append("EPSG:4326")
            gdal_translate_params.append(input_file)
            gdal_translate_params.append(cropped_input_file)

            p = Process(
                exec_type="exec",
                executable=gdal_translate,
                executable_params=gdal_translate_params,
                id=f"gdal_translate_{self.product_id}",
                skip_permission_check=True,
            )
            import_commands.append(p)

            p = Process(
                exec_type="grass",
                executable="r.import",
                executable_params=[
                    "input=%s" % cropped_input_file,
                    "output=%s" % temp_map_name,
                    "--q",
                ],
                id=f"r_import_{self.product_id}",
                skip_permission_check=True,
            )
            import_commands.append(p)

            p = Process(
                exec_type="grass",
                executable="g.region",
                executable_params=[
                    "align=%s" % temp_map_name,
                    "vector=%s" % self.product_id,
                    "-g",
                ],
                id=f"set_g_region_to_{self.product_id}",
                skip_permission_check=True,
            )
            import_commands.append(p)

            p = Process(
                exec_type="grass",
                executable="r.mask",
                executable_params=["vector=%s" % self.product_id],
                id=f"r_mask_{self.product_id}",
                skip_permission_check=True,
            )
            import_commands.append(p)

            p = Process(
                exec_type="grass",
                executable="r.mapcalc",
                executable_params=[
                    "expression=%s = float(%s)" % (map_name, temp_map_name)
                ],
                id=f"create_float_rastermap_{self.product_id}",
                skip_permission_check=True,
            )
            import_commands.append(p)

            p = Process(
                exec_type="grass",
                executable="r.timestamp",
                executable_params=["map=%s" % map_name, "date=%s" % timestamp],
                id=f"r_timestamp_{self.product_id}",
                skip_permission_check=True,
            )
            import_commands.append(p)

            p = Process(
                exec_type="grass",
                executable="g.remove",
                executable_params=[
                    "type=raster",
                    "name=%s" % temp_map_name,
                    "-f",
                ],
                id=f"remove_tmp_map_{self.product_id}",
                skip_permission_check=True,
            )
            import_commands.append(p)

            p = Process(
                exec_type="grass",
                executable="r.mask",
                executable_params=["-r"],
                id=f"remove_mask_{self.product_id}",
                skip_permission_check=True,
            )
            import_commands.append(p)

        return import_commands

    def get_ndvi_r_mapcalc_process_list(
        self, red, nir, raster_result_name="ndvi"
    ):
        """Create NDVI r.mapcalc command and set the color table accordingly

        NDVI formular: (nir - red) / (nir + red)

        Args:
            red (str): The map name of the red band
            nir (str): The map name of the near infrared band
            raster_result_name (str): The result raster map name

        Returns:
            ndvi_commands

        """

        ndvi_commands = []

        p = Process(
            exec_type="grass",
            executable="r.mapcalc",
            executable_params=[
                "expression=%(ndvi)s = (float(%(nir)s) - float(%(red)s))/"
                "(float(%(nir)s) + float(%(red)s))"
                % {"ndvi": raster_result_name, "nir": nir, "red": red}
            ],
            id=f"compute_NDVI_{raster_result_name}",
            skip_permission_check=True,
        )
        ndvi_commands.append(p)

        p = Process(
            exec_type="grass",
            executable="r.colors",
            executable_params=["color=ndvi", "map=%s" % raster_result_name],
            id=f"set_color_{raster_result_name}",
            skip_permission_check=True,
        )
        ndvi_commands.append(p)

        return ndvi_commands



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Base class for asynchronous processing
"""

import pickle
from flask import jsonify, make_response
from actinia_core.core.common.redis_interface import enqueue_job
from actinia_core.rest.base.resource_base import ResourceBase
from actinia_core.processing.common.ephemeral_processing import start_job

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Anika Weinmann"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class AsyncEphemeralResource(ResourceBase):
    """This class represents a processing resource that works on a temporary
    mapset.
    """

    def __init__(self, resource_id=None, iteration=None, post_url=None):
        ResourceBase.__init__(self, resource_id, iteration, post_url)

    def post(self, location_name):
        """Start an async GRASS processing task, that is completely temporary.
        The generated mapset is only created temporally, all created resources
        will be deleted after the processing finished.

        None:
            This resource should not be used in production, only for benchmarks
            and tests!

        Args:
            location_name (str): The name of the location

        Returns:
            flask.Response:
            The HTTP status and a JSON document that includes the
            status URL of the task that must be polled for updates.

        Process chain input format::

            {                                           # A process chain is a
                                                          dict with entries for
                                                          each module that
                                                          should be run
               Id:{                                     # Id must be numerical
                                                          and indicates the
                                                          process order
                    "module":<module_name>,             # Name of the module to
                                                          run
                    "stdin":<Id::stdout | Id::stderr>   # Use the output of a
                                                          specific module as
                                                          input for this module
                                                        # Id:stdout, Id:stderr
                                                          are available
                    "inputs":{                          # Definition of all
                                                          input parameters as
                                                          key:value pairs
                             <parameter name>:<value>,  # e.g.: value == "raste
                                                          r_name@mapset_name"
                                                          or "degree" or 0.0
                             <parameter name>:<value>   # e.g.: value == $file
                                                          ::slope_output_file
                                                          to specify an output
                                                          file
                                                        # that name will be
                                                          automatically
                                                          generated by the API.
                    },
                    "outputs":{                         # Definition of all
                                                          outputs using
                                                          key:value pairs
                        <parameter name>:{
                            "name":<value>,             # Output name e.g.
                                                          "my_aspect_map" or a
                                                          temporary file id
                                                        # definition: $file::id
                                                          eg: $file::aspect_
                                                          output_file
                                                        # This file can be used
                                                          in other module as
                                                          input
                            "export":{                  # Export options, if
                                                          present this map will
                                                          be exported
                                "format":<value>        # Set the export format
                                                          raster=GeoTiff
                                                          (default), vector =
                                                          shape (default)
                                "type":<output type>,   # Output type e.g.:
                                                          raster, vector, file,
                                                          stds
                            }
                        },
                        <parameter name>:{
                            "name":<value>,             # Output name e.g.
                                                          "my_slope_map"
                            "export":{                  # Export options, if
                                                          present this map will
                                                          be exported
                                "format":<value>        # Set the export format
                                                          raster=GeoTiff
                                                          (default), vector =
                                                          shape (default)
                                "type":<output type>,   # Output type e.g.:
                                                          raster, vector, file,
                                                          stds
                            }
                        }
                    },
                    "flags":<flags>,                    # All flags in a string
                                                          e.g.: "ge"
                    "overwrite":<True|False>,           # Set True to overwrite
                                                          existing data
                    "verbose":<True|False>              # Verbosity of the
                                                          module
                },
               Id:{                                     # The id of an
                                                          executable command,
                                                          that is not a grass
                                                          module
                    "executable":<path>,                # The name and path of
                                                          the executable e.g.
                                                          /bin/cp
                    "stdin":<Id::stdout | Id::stderr>   # Use the output of a
                                                          specific module as
                                                          input for this module
                                                        # Id::stdout, Id::
                                                          stderr are available
                    "parameters":[<parameter>,]         # A list of strings
                                                          that represent the
                                                          parameters that may
                                                          contain
                                                        # temporary file
                                                          definitions:
                                                          $file::id  eg:
                                                          $file::aspect_output
                                                          _file
               },
                ...
            }

        """
        # Preprocess the post call
        rdc = self.preprocess(location_name=location_name)

        if rdc:
            enqueue_job(self.job_timeout, start_job, rdc)

        html_code, response_model = pickle.loads(self.response_data)
        return make_response(jsonify(response_model), html_code)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Location management

TODO: Integrate into the ephemeral process chain approach
"""

from flask import jsonify, make_response
import os
import shutil
import pickle
from flask_restful_swagger_2 import swagger
from actinia_api.swagger2.actinia_core.apidocs import location_management
from actinia_api.swagger2.actinia_core.schemas.location_management import (
    LocationListResponseModel,
)

from actinia_core.core.common.app import auth
from actinia_core.core.common.api_logger import log_api_call
from actinia_core.rest.base.endpoint_config import (
    check_endpoint,
    endpoint_decorator,
)
from actinia_core.rest.base.user_auth import check_user_role
from actinia_core.rest.base.user_auth import check_user_permissions
from actinia_core.models.response_models import SimpleResponseModel
from actinia_core.rest.base.resource_base import ResourceBase
from actinia_core.core.common.redis_interface import enqueue_job
from actinia_core.core.utils import ensure_valid_path
from actinia_core.processing.common.location_management import (
    read_current_region,
    create_location,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Carmen Tawalika, Anika Weinmann"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class ListLocationsResource(ResourceBase):
    """This resource represents GRASS GIS database directory
    that contains locations.
    """

    def __init__(self):
        ResourceBase.__init__(self)

    """
    Return a list of all available locations that are located in the GRASS
    database
    """
    layer_type = None

    @endpoint_decorator()
    @swagger.doc(check_endpoint("get", location_management.get_doc))
    def get(self):
        """Get a list of all available locations"""
        locations = []

        if os.path.isdir(self.grass_data_base):
            dirs = os.listdir(self.grass_data_base)
            for dir in dirs:
                dir_path = os.path.join(self.grass_data_base, dir)
                if os.path.isdir(dir_path) and os.access(
                    dir_path, os.R_OK & os.X_OK
                ):
                    # Check for PERMANENT mapset existence
                    mapset_path = os.path.join(dir_path, "PERMANENT")
                    if os.path.isdir(mapset_path) and os.access(
                        mapset_path, os.R_OK & os.X_OK
                    ):
                        # Check access rights to the global database
                        # Super admin can see all locations
                        if (
                            self.has_superadmin_role
                            or dir
                            in self.user_credentials["permissions"][
                                "accessible_datasets"
                            ]
                        ):
                            locations.append(dir)
        # List all locations in the user database
        user_database = os.path.join(
            self.grass_user_data_base, self.user_group
        )
        if os.path.isdir(user_database):
            dirs = os.listdir(user_database)
            for dir in dirs:
                dir_path = os.path.join(user_database, dir)
                if os.path.isdir(dir_path) and os.access(
                    dir_path, os.R_OK & os.X_OK
                ):
                    # Check for PERMANENT mapset existence
                    mapset_path = os.path.join(dir_path, "PERMANENT")
                    if os.path.isdir(mapset_path) and os.access(
                        mapset_path, os.R_OK & os.X_OK
                    ):
                        locations.append(dir)
        if locations:
            return make_response(
                jsonify(
                    LocationListResponseModel(
                        status="success", locations=locations
                    )
                ),
                200,
            )
        else:
            return make_response(
                jsonify(
                    SimpleResponseModel(
                        status="error",
                        message="Unable to access GRASS database.",
                    )
                ),
                405,
            )


class LocationManagementResourceUser(ResourceBase):
    """This class returns information about a specific location"""

    def __init__(self):
        ResourceBase.__init__(self)

    @endpoint_decorator()
    @swagger.doc(check_endpoint("get", location_management.get_user_doc))
    def get(self, location_name):
        """
        Get the location projection and current computational region of the
        PERMANENT mapset
        """
        rdc = self.preprocess(
            has_json=False,
            has_xml=False,
            location_name=location_name,
            mapset_name="PERMANENT",
        )
        if rdc:
            enqueue_job(
                self.job_timeout,
                read_current_region,
                rdc,
                queue_type_overwrite=True,
            )
            http_code, response_model = self.wait_until_finish()
        else:
            http_code, response_model = pickle.loads(self.response_data)

        return make_response(jsonify(response_model), http_code)


class LocationManagementResourceAdminUser(ResourceBase):
    """This class manages the creation, deletion and modification of locations

    This is only allowed for administrators and users
    """

    decorators = [
        log_api_call,
        check_user_permissions,
        check_user_role,
        auth.login_required,
    ]

    def __init__(self):
        ResourceBase.__init__(self)

    @endpoint_decorator()
    @swagger.doc(check_endpoint("delete", location_management.delete_user_doc))
    def delete(self, location_name):
        """
        Delete an existing location and everything inside from the user
        database.
        """
        # Delete only locations from the user database
        location = ensure_valid_path(
            [self.grass_user_data_base, self.user_group, location_name]
        )
        permanent_mapset = ensure_valid_path([location, "PERMANENT"])
        wind_file = ensure_valid_path([permanent_mapset, "WIND"])
        # Check the location path, only "valid" locations can be deleted
        if os.path.isdir(location):
            if os.path.isdir(permanent_mapset) and os.path.isfile(wind_file):
                try:
                    shutil.rmtree(location)
                    return make_response(
                        jsonify(
                            SimpleResponseModel(
                                status="success",
                                message="location %s deleted" % location_name,
                            )
                        ),
                        200,
                    )
                except Exception as e:
                    return make_response(
                        jsonify(
                            SimpleResponseModel(
                                status="error",
                                message="Unable to delete location "
                                f"{location_name} Exception {e}",
                            )
                        ),
                        500,
                    )

        return make_response(
            jsonify(
                SimpleResponseModel(
                    status="error",
                    message="location %s does not exists" % location_name,
                )
            ),
            400,
        )

    @endpoint_decorator()
    @swagger.doc(check_endpoint("post", location_management.post_user_doc))
    def post(self, location_name):
        """Create a new location based on EPSG code in the user database."""
        # Create only new locations if they did not exist in the global
        # database
        location = ensure_valid_path([self.grass_data_base, location_name])

        # Check the location path
        if os.path.isdir(location):
            return self.get_error_response(
                message="Unable to create location. "
                "Location <%s> exists in global database." % location_name
            )

        # Check also for the user database
        location = ensure_valid_path(
            [self.grass_user_data_base, self.user_group, location_name]
        )
        # Check the location path
        if os.path.isdir(location):
            return self.get_error_response(
                message="Unable to create location. "
                "Location <%s> exists in user database." % location_name
            )

        rdc = self.preprocess(
            has_json=True,
            has_xml=False,
            location_name=location_name,
            mapset_name="PERMANENT",
        )
        if rdc:
            enqueue_job(self.job_timeout, create_location, rdc)
            http_code, response_model = self.wait_until_finish()
        else:
            http_code, response_model = pickle.loads(self.response_data)

        return make_response(jsonify(response_model), http_code)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Storage base class
"""
import os
import datetime
from google.cloud import storage
from .storage_interface_base import ResourceStorageBase

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class ResourceStorageGCS(ResourceStorageBase):
    """
    Storage class of generated resources to be put in a Google Cloud Storage
    bucket
    """

    def __init__(self, user_id, resource_id, config):
        """
        Storage class of generated resources to be put in a Google Cloud
        Storage bucket

        Args:
            user_id: The user id
            resource_id: The resource id
            config: The configuration of Actinia Core
        """
        ResourceStorageBase.__init__(self, user_id, resource_id, config)

        self.storage_client = None
        self.bucket_name = self.config.GCS_RESOURCE_BUCKET

    def setup(self):
        """
        Setup the Google Cloud Storage (GCS) client and the GCS credentials
        """
        os.environ[
            "GOOGLE_APPLICATION_CREDENTIALS"
        ] = self.config.GOOGLE_APPLICATION_CREDENTIALS
        self.storage_client = storage.Client()

    def get_resource_urls(self):
        """Return all resource urls that were generated

        Returns:
            (list): A list of urls

        """
        return self.resource_url_list

    def store_resource(self, file_path):
        """Store a resource (file) at the user resource storage and return an
        URL to the resource accessible via HTTP(S)

        Args:
            file_path:

        Returns:
            (str): the resource url that points to the stored resource
        """

        file_name = os.path.basename(file_path)
        object_path = os.path.join(self.user_id, self.resource_id, file_name)

        if not self.bucket_name:
            Exception("No storage bucket was defined")

        bucket = self.storage_client.get_bucket(self.bucket_name)
        blob = bucket.blob(object_path)
        blob.upload_from_filename(file_path)

        # Generate a persistent URL from the Bucket
        url = blob.generate_signed_url(
            # This URL is valid for 10 days
            expiration=datetime.timedelta(days=10),
            # Allow GET requests using this URL.
            method="GET",
        )

        self.resource_file_list.append(object_path)
        self.resource_url_list.append(url)
        return url

    def remove_resources(self):
        """Remove the resource export path and everything inside"""
        for blob_name in self.resource_file_list:
            bucket = self.storage_client.get_bucket(self.bucket_name)
            blob = bucket.blob(blob_name)
            blob.delete()



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""The user specific resources

This module specifies all endpoints to manage user accounts
in the redis database via REST API calls.

TODO: Implement POST full permission creation
      Implement PUT to modify existing users
"""

from flask import jsonify, make_response, g
from flask_restful import reqparse
from flask_restful_swagger_2 import swagger
from actinia_api.swagger2.actinia_core.apidocs import user_management

from actinia_core.core.common.config import global_config
from actinia_core.rest.base.endpoint_config import (
    check_endpoint,
    endpoint_decorator,
)
from actinia_core.rest.base.base_login import LoginBase
from actinia_core.rest.base.user_auth import (
    check_admin_role,
    check_admin_role_or_own_userid,
)
from actinia_core.core.common.api_logger import log_api_call
from actinia_core.core.common.app import auth
from actinia_core.core.common.user import ActiniaUser
from actinia_core.models.response_models import (
    UserListResponseModel,
    UserInfoResponseModel,
    SimpleResponseModel,
)


__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class UserListResource(LoginBase):
    """List all user in the database."""

    def __init__(self):
        LoginBase.__init__(self)

    @endpoint_decorator()
    @swagger.doc(check_endpoint("get", user_management.user_list_get_doc))
    def get(self):
        """List all users in the database

        These methods work only if the
        authorized user has an admin role.

        Returns:
            flask.Response: A HTTP response with
                            JSON payload containing a list of users
        """
        if global_config.KEYCLOAK_CONFIG_PATH:
            return make_response(
                jsonify(
                    SimpleResponseModel(
                        status="error",
                        message="The keycloak authentication does not allow "
                        "to request all users",
                    )
                ),
                400,
            )

        user = ActiniaUser(None)
        user_list = user.list_all_users()

        return make_response(
            jsonify(
                UserListResponseModel(status="success", user_list=user_list)
            ),
            200,
        )


class UserManagementResource(LoginBase):
    """Get, Create and Delete a single user

    These methods work only if the
    authorized user has an admin role.

    Only normal users (role=user) can be created with this class.

    """

    # Authorization is required for all resources
    # API logging is required for all resources
    decorators = [log_api_call, auth.login_required]

    @endpoint_decorator()
    @swagger.doc(check_endpoint("get", user_management.user_get_doc))
    @check_admin_role_or_own_userid
    def get(self, user_id):
        """Return the credentials of a single user

        These methods work only if the
        authorized user has an admin role or requests the own user ID.

        Args:
            user_id (str): The unique name of the user

        Returns:
            flask.Response: A HTTP response with
                            JSON payload containing the credentials
                            of the user
        """
        if global_config.KEYCLOAK_CONFIG_PATH:
            user = g.user
            if user.user_id != user_id:
                return make_response(
                    jsonify(
                        SimpleResponseModel(
                            status="error",
                            message="The keycloak authentication does not "
                            "allow to request another user.",
                        )
                    ),
                    400,
                )
        else:
            user = ActiniaUser(user_id)
            if user.exists() != 1:
                return make_response(
                    jsonify(
                        SimpleResponseModel(
                            status="error",
                            message="User <%s> does not exist" % user_id,
                        )
                    ),
                    400,
                )

        credentials = user.get_credentials()

        return make_response(
            jsonify(
                UserInfoResponseModel(
                    status="success",
                    permissions=credentials["permissions"],
                    user_id=credentials["user_id"],
                    user_role=credentials["user_role"],
                    user_group=credentials["user_group"],
                )
            ),
            200,
        )

    @endpoint_decorator()
    @swagger.doc(check_endpoint("post", user_management.user_post_doc))
    @check_admin_role
    def post(self, user_id):
        """Create a user in the database

        These methods work only if the
        authorized user has an admin role.

        Args:
            user_id (str): The unique name of the user

        Returns:
            flask.Response: A HTTP response with
                            JSON payload containing
                            the status and messages
        """
        if global_config.KEYCLOAK_CONFIG_PATH:
            return make_response(
                jsonify(
                    SimpleResponseModel(
                        status="error",
                        message="The keycloak authentication does not allow "
                        "to create a new user",
                    )
                ),
                400,
            )

        user = ActiniaUser(user_id)

        if user.exists() == 1:
            return make_response(
                jsonify(
                    SimpleResponseModel(
                        status="error",
                        message="User <%s> already exists" % user_id,
                    )
                ),
                400,
            )

        # Password parser
        password_parser = reqparse.RequestParser()
        password_parser.add_argument(
            "password",
            required=True,
            type=str,
            location="args",
            dest="password",
            help="The password of the new user cannot be converted.",
        )
        password_parser.add_argument(
            "group",
            required=True,
            type=str,
            location="args",
            dest="group",
            help="The group of the new user cannot be converted.",
        )
        args = password_parser.parse_args()
        password = args["password"]
        group = args["group"]

        user = ActiniaUser.create_user(user_id, group, password, "user", {})
        if user is not None:
            if user.exists():
                return make_response(
                    jsonify(
                        SimpleResponseModel(
                            status="success",
                            message="User %s created" % user_id,
                        )
                    ),
                    201,
                )

        return make_response(
            jsonify(
                SimpleResponseModel(
                    status="error",
                    message="Unable to create user %s" % user_id,
                )
            ),
            400,
        )

    @endpoint_decorator()
    @swagger.doc(check_endpoint("delete", user_management.user_delete_doc))
    @check_admin_role
    def delete(self, user_id):
        """Delete a specific user

        These methods work only if the
        authorized user has an admin role.

        Args:
            user_id (str): The unique name of the user

        Returns:
            flask.Response: A HTTP response with
                            JSON payload containing
                            the status and messages
        """
        if global_config.KEYCLOAK_CONFIG_PATH:
            return make_response(
                jsonify(
                    SimpleResponseModel(
                        status="error",
                        message="The keycloak authentication does not allow "
                        "to delete a user",
                    )
                ),
                400,
            )

        user = ActiniaUser(user_id)

        if user.exists() != 1:
            return make_response(
                jsonify(
                    SimpleResponseModel(
                        status="error",
                        message=f"Unable to delete user {user_id}. User does "
                        "not exist.",
                    )
                ),
                400,
            )

        if user.delete() is True:
            return make_response(
                jsonify(
                    SimpleResponseModel(
                        status="success", message="User %s deleted" % user_id
                    )
                ),
                200,
            )

        return make_response(
            jsonify(
                SimpleResponseModel(
                    status="error",
                    message="Unable to delete user %s" % user_id,
                )
            ),
            400,
        )



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2021 mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
This code shows the STACImporter model which helps actinia core
to save and manage the STAC collections stored through the actinia STAC plugin
"""

# from PyQt4.QtCore import *
# from PyQt4.QtXml import *
# from config import global_config

__license__ = "GPLv3"
__author__ = "Jorge Herrera"
__copyright__ = "Copyright 2021, mundialis GmbH & Co. KG"
__maintainer__ = "__mundialis__"
__email__ = "info@mundialis.de"


import requests
import os
import json
from actinia_core.core.common.exceptions import AsyncProcessError
from actinia_core.core.common.process_object import Process

try:
    from actinia_stac_plugin.core.stac_collection_id import callStacCollection

    has_plugin = True
except Exception:
    has_plugin = False


class STACImporter:
    @staticmethod
    def _get_search_root(stac_collection_id):
        stac_from_actinia = callStacCollection(stac_collection_id)
        try:
            stac_json = json.loads(stac_from_actinia)
        except Exception:
            stac_json = stac_from_actinia

        for item in stac_json["links"]:
            if item["rel"] == "root":
                stac_url = item["href"]

                if "?" in stac_url:
                    stac_url = stac_url.split("?")[0]

                if stac_url.endswith("/"):
                    stac_root_search = stac_url + "search"
                else:
                    stac_root_search = stac_url + "/search"

        return stac_root_search

    @staticmethod
    def _apply_filter(stac_root_search, stac_name, interval, bbox, filter):
        search_body = {
            "collections": [stac_name],
        }
        search_body["query"] = filter

        search_body["bbox"] = bbox

        search_body["interval"] = interval

        stac_search = requests.post(stac_root_search, json=search_body)

        full_filtered_result = stac_search.json()

        if (
            "features" in full_filtered_result
            and len(full_filtered_result["features"]) > 0
        ):
            return full_filtered_result
        else:
            stac_search = requests.get(stac_root_search, json=search_body)
            full_filtered_result = stac_search.json()

            if (
                "features" in full_filtered_result
                and len(full_filtered_result["features"]) > 0
            ):
                return full_filtered_result
            else:
                raise AsyncProcessError("Not matched found")

    @staticmethod
    def _get_filtered_bands(stac_items, semantic_label):
        band_roots = {}

        for feature in stac_items["features"]:
            item_date = feature["properties"]["datetime"]
            for key, value in feature["assets"].items():
                if "eo:bands" in value:
                    if "common_name" in value["eo:bands"][0]:
                        if (
                            value["eo:bands"][0]["common_name"]
                            in semantic_label
                            or value["eo:bands"][0]["name"] in semantic_label
                            or semantic_label == []
                        ):
                            band_name = value["eo:bands"][0]["name"]
                            if band_name not in band_roots:
                                band_roots[band_name] = {}
                            feature_id = feature["id"]
                            item_link = feature["assets"][band_name]["href"]
                            band_roots[band_name]["name_id"] = feature_id
                            band_roots[band_name]["url"] = item_link
                            band_roots[band_name]["datetime"] = item_date
        return band_roots

    def _stac_import(
        self,
        stac_collection_id=None,
        semantic_label=None,
        interval=None,
        bbox=None,
        filter=None,
        strd_name=None,
    ):
        if has_plugin:
            try:
                stac_name = stac_collection_id.split(".")[3]
            except Exception:
                raise AsyncProcessError(
                    "The source has not the right structure"
                )

            stac_root = self._get_search_root(stac_collection_id)

            stac_filtered = self._apply_filter(
                stac_root, stac_name, interval, bbox, filter
            )

            stac_result = self._get_filtered_bands(
                stac_filtered, semantic_label
            )

            stac_processes = []

            # Create the strds

            # dd/mm/YY H:M:S

            exec_params = [
                "type=strds",
                "temporaltype=absolute",
                "output=%s" % strd_name,
                "title=%s" % strd_name,
                "description=%s" % f"{stac_collection_id}",
            ]

            p = Process(
                exec_type="grass",
                executable="t.create",
                executable_params=exec_params,
                id=f"t_create_{os.path.basename(stac_name)}",
                skip_permission_check=True,
            )

            stac_processes.append(p)

            for key, value in stac_result.items():
                item_id = value["name_id"]
                output_name = f"{strd_name}_{item_id}_{key}"

                url_prefix = "/vsicurl/"
                # Checking if the URL belongs to S3
                if "s3:" in value["url"]:
                    url_prefix = "/vsis3/"

                # Upload the image to GRASS
                exec_params = [
                    "input=%s" % url_prefix + value["url"],
                    "output=%s" % output_name,
                    "extent=region",
                ]

                import_raster = Process(
                    exec_type="grass",
                    executable="r.import",
                    executable_params=exec_params,
                    id=f"r_import_{output_name}",
                    skip_permission_check=True,
                )

                stac_processes.append(import_raster)

                # Setting the Semantic Label
                exec_params_sl = [
                    "map=%s" % output_name,
                    "semantic_label=%s" % key,
                ]

                sem_lab = Process(
                    exec_type="grass",
                    executable="r.support",
                    executable_params=exec_params_sl,
                    id=f"r_semantic_label_{output_name}",
                    skip_permission_check=True,
                )

                stac_processes.append(sem_lab)

                # Register the raster to the STDR
                exec_params_stdr = [
                    "input=%s" % strd_name,
                    "type=raster",
                    "maps=%s" % output_name,
                    "start=%s" % value["datetime"],
                ]

                registration = Process(
                    exec_type="grass",
                    executable="t.register",
                    executable_params=exec_params_stdr,
                    id=f"t_register_{output_name}",
                    skip_permission_check=True,
                )

                stac_processes.append(registration)
        else:
            raise AsyncProcessError("Actinia STAC plugin is not installed")

        return stac_processes

    def get_stac_import_download_commands(
        self,
        stac_entry,
        config=None,
        temp_file_path=None,
        message_logger=None,
        send_resource_update=None,
    ):
        """Helper method to get the stac import and download commands.
        Args:
            stac_entry (dict): stac_entry of the import description list
        Returns:
            stac_commands: The stac download and import commands
        """
        # Check for band information
        # TODO check config, temp_file_path, message_logger,
        # send_resource_update
        stac_entry_source = stac_entry["import_descr"]["source"]

        if "semantic_label" in stac_entry["import_descr"]:
            stac_semantic_label = stac_entry["import_descr"]["semantic_label"]

        if "extent" in stac_entry["import_descr"]:
            if (
                "spatial"
                and "temporal" not in stac_entry["import_descr"]["extent"]
            ):
                raise AsyncProcessError(
                    "Unknown spatial or/and temporal parameters"
                    "in the process chain definition"
                )

            if "bbox" in stac_entry["import_descr"]["extent"]["spatial"]:
                stac_extent = stac_entry["import_descr"]["extent"]["spatial"][
                    "bbox"
                ][0]

            if "interval" in stac_entry["import_descr"]["extent"]["temporal"]:
                interval = stac_entry["import_descr"]["extent"]
                interval = interval["temporal"]["interval"][0]
                stac_interval = interval

        stac_filter = {}
        if "filter" in stac_entry["import_descr"]:
            stac_filter = stac_entry["import_descr"]["filter"]

        stac_name = stac_entry["value"]

        stac_command = self._stac_import(
            stac_collection_id=stac_entry_source,
            semantic_label=stac_semantic_label,
            interval=stac_interval,
            bbox=stac_extent,
            filter=stac_filter,
            strd_name=stac_name,
        )
        return stac_command



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
STRDS map layer management

TODO: Integrate into the ephemeral process chain approach
"""
from actinia_api.swagger2.actinia_core.schemas.strds_management import (
    STRDSInfoModel,
    STRDSInfoResponseModel,
)

from actinia_core.processing.actinia_processing.ephemeral.persistent_processing import (
    PersistentProcessing,
)
from actinia_core.core.common.exceptions import AsyncProcessError
from actinia_core.models.response_models import (
    StringListProcessingResultResponseModel,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert, Carmen Tawalika"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class PersistentSTRDSLister(PersistentProcessing):
    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)
        self.response_model_class = StringListProcessingResultResponseModel

    def _execute(self):
        self._setup()

        pc = {
            "1": {
                "module": "t.list",
                "inputs": {"type": "strds", "column": "name"},
            }
        }

        # Make sure that only the current mapset is used for strds listing
        has_where = False

        if self.rdc.user_data:
            for option in self.rdc.user_data:
                if self.rdc.user_data[option] is not None:
                    if "where" in option:
                        select = (
                            self.rdc.user_data[option]
                            + " AND mapset = '%s'" % self.mapset_name
                        )
                        pc["1"]["inputs"]["where"] = select
                        has_where = True
                    else:
                        pc["1"]["inputs"][option] = self.rdc.user_data[option]

        if has_where is False:
            select = "mapset='%s'" % self.mapset_name
            pc["1"]["inputs"]["where"] = select

        process_list = self._validate_process_chain(
            skip_permission_check=True, process_chain=pc
        )
        self._create_temp_database()
        self._create_grass_environment(
            grass_data_base=self.temp_grass_data_base,
            mapset_name=self.mapset_name,
        )

        self._execute_process_list(process_list)

        mapset_lists = []
        mapsets = self.module_output_log[0]["stdout"].split()

        for mapset in mapsets:
            mapset_lists.append(mapset.strip())

        self.module_results = mapset_lists


class PersistentSTRDSInfo(PersistentProcessing):
    """Gather the STRDS information"""

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)
        self.response_model_class = STRDSInfoResponseModel

    def _execute(self):
        self._setup()

        pc = {
            "1": {
                "module": "t.info",
                "inputs": {"type": "strds", "input": self.map_name},
                "flags": "g",
            }
        }

        process_list = self._validate_process_chain(
            skip_permission_check=True, process_chain=pc
        )
        self._create_temp_database()
        self._create_grass_environment(
            grass_data_base=self.temp_grass_data_base,
            mapset_name=self.mapset_name,
        )

        self._execute_process_list(process_list)

        kv_list = self.module_output_log[0]["stdout"].split("\n")

        strds = {}

        for string in kv_list:
            if "=" in string:
                k, v = string.split("=", 1)
                strds[k] = v

        self.module_results = STRDSInfoModel(**strds)
        self.finish_message = (
            "Information gathering for STRDS <%s> successful" % self.map_name
        )


class PersistentSTRDSDeleter(PersistentProcessing):
    """Delete a STRDS"""

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)

    def _execute(self):
        self._setup()
        self.required_mapsets.append(self.target_mapset_name)

        args = self.rdc.user_data

        pc = {
            "1": {
                "module": "t.remove",
                "inputs": {"type": "strds", "inputs": self.map_name},
                "flags": "f",
            }
        }

        if args and "recursive" in args and args["recursive"] is True:
            pc["1"]["flags"] = "rf"

        process_list = self._validate_process_chain(
            skip_permission_check=True, process_chain=pc
        )

        self._create_temp_database()
        self._check_lock_target_mapset()

        self._create_grass_environment(
            grass_data_base=self.temp_grass_data_base,
            mapset_name=self.target_mapset_name,
        )

        self._execute_process_list(process_list)
        self.finish_message = "STRDS <%s> successfully deleted" % self.map_name


class PersistentSTRDSCreator(PersistentProcessing):
    """Create a STRDS"""

    def __init__(self, *args):
        PersistentProcessing.__init__(self, *args)

    def _execute(self):
        self._setup()
        self.required_mapsets.append(self.target_mapset_name)

        pc_1 = {}
        pc_1["1"] = {
            "module": "t.list",
            "inputs": {
                "type": "strds",
                "where": "id = '%s@%s'"
                % (self.map_name, self.target_mapset_name),
            },
        }
        # Check the first process chain
        pc_1 = self._validate_process_chain(
            skip_permission_check=True, process_chain=pc_1
        )

        pc_2 = {
            "1": {
                "module": "t.create",
                "inputs": {"type": "strds", "output": self.map_name},
            }
        }

        if self.request_data:
            for key in self.request_data:
                pc_2["1"]["inputs"][key] = self.request_data[key]

        pc_2 = self._validate_process_chain(
            skip_permission_check=True, process_chain=pc_2
        )
        self._create_temp_database()
        self._check_lock_target_mapset()

        self._create_grass_environment(
            grass_data_base=self.temp_grass_data_base,
            mapset_name=self.target_mapset_name,
        )

        self._execute_process_list(pc_1)

        # check if STRDS exists
        raster_list = self.module_output_log[0]["stdout"].split("\n")

        if len(raster_list[0]) > 0:
            raise AsyncProcessError("STRDS <%s> exists." % self.map_name)

        self._execute_process_list(pc_2)

        self.finish_message = "STRDS <%s> successfully created" % self.map_name



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2018 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Storage base class
"""
import os
import shutil
from .storage_interface_base import ResourceStorageBase

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2018, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "Sören Gebbert"
__email__ = "soerengebbert@googlemail.com"


class ResourceStorageFilesystem(ResourceStorageBase):
    """Storage class of generated resources at the disk storage that is
    configured in the config file
    """

    def __init__(self, user_id, resource_id, config, resource_url_base=None):
        """Storage class of generated resources at the disk storage that is
        configured in the config file

        Args:
            user_id: The user id
            resource_id: The resource id
            config: The configuration of Actinia Core
            resource_url_base: The base URL name that must be used to create
                               the resource urls
        """
        ResourceStorageBase.__init__(self, user_id, resource_id, config)

        self.resource_url_base = resource_url_base
        self.resource_dir = self.config.GRASS_RESOURCE_DIR
        self.user_export_path = os.path.join(self.resource_dir, self.user_id)
        self.resource_export_path = os.path.join(
            self.user_export_path, self.resource_id
        )
        self.dir_created = False

    def _create_resource_directory(self):
        """Create a new resource directory based on the resource base, the user
        id and the resource id. This directory will contain all exported files
        of the process chain.

        This method will create the directory only once and will ignore
        repeating calls.

        Raises:
            IOError: If the resource path is not accessible.
        """

        if self.dir_created is True:
            return

        # check if the resource base is accessible
        if os.path.isdir(self.resource_dir) and os.access(
            self.resource_dir, os.R_OK | os.X_OK | os.W_OK
        ):
            if os.path.exists(self.user_export_path) is False:
                os.mkdir(self.user_export_path)
                os.chmod(self.user_export_path, 0o0700)

            # Check if the user path can be created if it does not exist
            if os.path.isdir(self.user_export_path) and os.access(
                self.user_export_path, os.R_OK | os.X_OK | os.W_OK
            ):
                # Create the resource directory
                if os.path.exists(self.resource_export_path) is False:
                    os.mkdir(self.resource_export_path)
                    os.chmod(self.resource_export_path, 0o0700)

                # Check access just in case the resource exists
                if os.path.isdir(self.resource_export_path) and os.access(
                    self.resource_export_path, os.R_OK | os.X_OK | os.W_OK
                ):
                    # TODO Add the directory to the scheduler list
                    pass
                else:
                    raise IOError(
                        "Unable to create resource directory <%s>"
                        % self.resource_export_path
                    )
            else:
                raise IOError(
                    "Unable to create resource user directory <%s>"
                    % self.user_export_path
                )
        else:
            raise IOError(
                "Unable to access the base resource directory <%s>"
                % self.resource_dir
            )

        self.dir_created = True

    def setup(self):
        """
        Setup the resource directory to store files in it via store_resource()

        Raises:
            IOError: If the resource path is not accessible.
        """
        self._create_resource_directory()

    def get_resource_urls(self):
        """
        Return all resource urls that were generated when storing a resource
        on disk

        Returns:
            (list): A list of urls

        """
        return self.resource_url_list

    def store_resource(self, file_path):
        """Store a resource (file) at the user resource storage and return an
        URL to the resource accessible via HTTP

        Args:
            file_path:

        Raises:
            IOError: If the resource path is not accessible.

        Returns:
            (str): the resource url that points to the stored resource

        """
        file_name = os.path.basename(file_path)
        export_path = os.path.join(self.resource_export_path, file_name)
        shutil.move(file_path, export_path)
        url = self.resource_url_base.replace("__None__", file_name)

        self.resource_url_list.append(url)
        self.resource_file_list.append(file_path)

        return url

    def remove_resources(self):
        """Remove the resource export path and everything inside"""
        if os.path.exists(self.resource_export_path) and os.path.isdir(
            self.resource_export_path
        ):
            shutil.rmtree(self.resource_export_path, ignore_errors=True)



# -*- coding: utf-8 -*-
#######
# actinia-core - an open source REST API for scalable, distributed, high
# performance processing of geographical data that uses GRASS GIS for
# computational tasks. For details, see https://actinia.mundialis.de/
#
# Copyright (c) 2016-2022 Sören Gebbert and mundialis GmbH & Co. KG
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#######

"""
Raster map renderer

"""
from flask_restful_swagger_2 import swagger
import os
from flask import jsonify, make_response, Response
from actinia_api.swagger2.actinia_core.apidocs import raster_renderer

from actinia_core.rest.base.endpoint_config import (
    check_endpoint,
    endpoint_decorator,
)
from actinia_core.core.common.redis_interface import enqueue_job
from actinia_core.rest.base.renderer_base import RendererBaseResource
from actinia_core.processing.common.raster_renderer import (
    start_job,
    start_rgb_job,
    start_shade_job,
)

__license__ = "GPLv3"
__author__ = "Sören Gebbert"
__copyright__ = (
    "Copyright 2016-2022, Sören Gebbert and mundialis GmbH & Co. KG"
)
__maintainer__ = "mundialis"


class SyncEphemeralRasterRendererResource(RendererBaseResource):
    """Render a raster image with g.region/d.rast approach synchronously"""

    @endpoint_decorator()
    @swagger.doc(check_endpoint("get", raster_renderer.raster_render_get_doc))
    def get(self, location_name, mapset_name, raster_name):
        """Render a raster map layer as a PNG image."""
        parser = self.create_parser()
        args = parser.parse_args()
        options = self.create_parser_options(args)

        if isinstance(options, dict) is False:
            return options

        rdc = self.preprocess(
            has_json=False,
            has_xml=False,
            location_name=location_name,
            mapset_name=mapset_name,
            map_name=raster_name,
        )

        rdc.set_user_data(options)

        enqueue_job(
            self.job_timeout, start_job, rdc, queue_type_overwrite=True
        )

        http_code, response_model = self.wait_until_finish(0.05)
        if http_code == 200:
            result_file = response_model["process_results"]
            # Open the image file, read it and then delete it
            if result_file:
                if os.path.isfile(result_file):
                    image = open(result_file, "rb").read()
                    os.remove(result_file)
                    return Response(image, mimetype="image/png")
        return make_response(jsonify(response_model), http_code)


class SyncEphemeralRasterRGBRendererResource(RendererBaseResource):
    """
    Render three raster layer as rgb image with g.region/d.rgb approach
    synchronously
    """

    def extract_rgb_maps(self, args, mapset_name):
        """Parse the request parameters for red green and blue raster layers

        Args:
            args (dict): The argument dictionary with rendering options
            mapset_name (str): Name of the mapset

        Returns:
             dict:
             A dictionary of d.rgb options or a error response if width or
             height are invalid and if the mapset is present in the raster
             layer names.

        """
        options = self.create_parser_options(args)

        if "@" in args["red"]:
            return self.get_error_response(
                message="Mapset name is not allowed in layer names"
            )

        if "@" in args["green"]:
            return self.get_error_response(
                message="Mapset name is not allowed in layer names"
            )

        if "@" in args["blue"]:
            return self.get_error_response(
                message="Mapset name is not allowed in layer names"
            )

        options["red"] = "%s@%s" % (args["red"], mapset_name)
        options["green"] = "%s@%s" % (args["green"], mapset_name)
        options["blue"] = "%s@%s" % (args["blue"], mapset_name)

        return options

    @endpoint_decorator()
    @swagger.doc(
        check_endpoint("get", raster_renderer.raster_rgb_render_get_doc)
    )
    def get(self, location_name, mapset_name):
        """Render three raster map layer as composed RGB PNG image."""

        parser = self.create_parser()
        parser.add_argument(
            "red",
            required=True,
            type=str,
            location="args",
            help="The name of the raster layer associated with the color red",
        )
        parser.add_argument(
            "green",
            required=True,
            type=str,
            location="args",
            help="The name of the raster layer associated with the color "
            "green",
        )
        parser.add_argument(
            "blue",
            required=True,
            type=str,
            location="args",
            help="The name of the raster layer associated with the color blue",
        )

        args = parser.parse_args()
        rgb_options = self.extract_rgb_maps(args, mapset_name)

        if isinstance(rgb_options, dict) is False:
            return rgb_options

        rdc = self.preprocess(
            has_json=False,
            has_xml=False,
            location_name=location_name,
            mapset_name=mapset_name,
        )

        rdc.set_user_data(rgb_options)

        enqueue_job(
            self.job_timeout, start_rgb_job, rdc, queue_type_overwrite=True
        )

        http_code, response_model = self.wait_until_finish(0.05)
        if http_code == 200:
            result_file = response_model["process_results"]
            # Open the image file, read it and then delete it
            if result_file:
                if os.path.isfile(result_file):
                    image = open(result_file, "rb").read()
                    os.remove(result_file)
                    return Response(image, mimetype="image/png")
        return make_response(jsonify(response_model), http_code)


class SyncEphemeralRasterShapeRendererResource(RendererBaseResource):
    """Render two raster layer as shade image with g.region/d.shade approach
    synchronously
    """

    def extract_shade_maps(self, args, mapset_name):
        """Parse the request parameters for shade and color raster layers

        Args:
            args (dict): The argument dictionary with rendering options
            mapset_name: Name of the mapset

        Returns:
             dict:
             A dictionary of d.shade options an error response if width or
             height are invalid and if the mapset is present in the raster
             layer names

        """
        options = self.create_parser_options(args)

        if "@" in args["shade"]:
            return self.get_error_response(
                message="Mapset name is not allowed in layer names"
            )

        if "@" in args["color"]:
            return self.get_error_response(
                message="Mapset name is not allowed in layer names"
            )

        options["shade"] = "%s@%s" % (args["shade"], mapset_name)
        options["color"] = "%s@%s" % (args["color"], mapset_name)

        return options

    @endpoint_decorator()
    @swagger.doc(
        check_endpoint("get", raster_renderer.raster_shade_render_get_doc)
    )
    def get(self, location_name, mapset_name):
        """Render two raster layers as a composed shade PNG image"""
        parser = self.create_parser()
        parser.add_argument(
            "shade",
            required=True,
            type=str,
            location="args",
            help="The name of the raster layer associated with shading",
        )
        parser.add_argument(
            "color",
            required=True,
            type=str,
            location="args",
            help="The name of the raster layer associated with color",
        )

        args = parser.parse_args()
        options = self.extract_shade_maps(args, mapset_name)

        if isinstance(options, dict) is False:
            return options

        rdc = self.preprocess(
            has_json=False,
            has_xml=False,
            location_name=location_name,
            mapset_name=mapset_name,
        )

        rdc.set_user_data(options)

        enqueue_job(
            self.job_timeout, start_shade_job, rdc, queue_type_overwrite=True
        )

        http_code, response_model = self.wait_until_finish(0.05)
        if http_code == 200:
            result_file = response_model["process_results"]
            # Open the image file, read it and then delete it
            if result_file:
                if os.path.isfile(result_file):
                    image = open(result_file, "rb").read()
                    os.remove(result_file)
                    return Response(image, mimetype="image/png")
        return make_response(jsonify(response_model), http_code)
